Set Implicit Arguments.
Set Contextual Implicit.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.
Require Import Relation_Definitions.
Require Import TLC.LibLN.
Require Import Arith.
Require Import ImpAlg.Lattices.
Require Import ImpAlg.Ens.
Require Import ImpAlg.ImpAlg.
Require Import ImpAlg.BotTop.


Generalizable Variables X ord join meet arrow L meet_set join_set.



Section ImpStructure_lambda.
  Context `{IS:ImpStructure}.


  (******************************************)
  (*        Encoding the λ-calculus         *)
  (******************************************)
  
  
  (* * We begin by defining the application of "a" to "b", 
       denoted by "a@b". *)
       
  
  Definition app a b:= ⊓ (fun x => a ≤ b ↦ x).
  Infix "@" := app (at level 59, left associativity).
  
  (* * Properties of application *)
  
  Lemma app_mon_l:∀ a b c, a ≤ b → a @ c ≤ b @ c.
  Proof.
    intros a b c Hab.
    unfold app.
    apply meet_intro.
    intros x Hx.
    apply meet_elim.
    apply (ord_trans _ _ _  Hab Hx).
  Qed.

  Lemma app_mon_r:∀ a b c, b ≤ c → a @ b ≤ a @ c.
  Proof.
    intros a b c Hbc.
    unfold app.
    apply meet_intro.
    intros x Hx.
    apply meet_elim.
    apply (ord_trans _ _ _  Hx).
    now apply arrow_mon_l.
  Qed.
  
  Lemma app_mon:∀ a a' b b', a≤ a' → b ≤ b' → a @ b ≤ a' @ b'.
  Proof.
    intros a a' b b' Ha Hb.
    eapply ord_trans;[apply (app_mon_l _ _ _ Ha)|].
    apply (app_mon_r _ _ _ Hb).
  Qed.
  
  
  Lemma app_beta: ∀ a b, (a ↦ b) @ a ≤ b.
  Proof.
    intros a b.
    unfold app.
    apply meet_elim.
    reflexivity.
  Qed.

  Lemma app_eta: ∀ a b, a ≤ b ↦ (a@b).
  Proof.
    intros a b. unfold app.
    rewrite arrow_meet.
    apply meet_intro.
    intros x (c,(Hx,Hc)). 
    now subst x.
  Qed.
  
  Lemma minimum: ∀ a b, min (fun x => a ≤ b ↦ x) (a@b).
  Proof.
  intros a b.
  split.
  - unfold lb,app. intros y Hy. now apply meet_elim. 
  - apply app_eta. 
  Qed.
  
  Lemma adjunction: ∀ a b c, a @ b ≤ c ↔ a ≤ b ↦ c.
  Proof.
    intros.
    split;intros.
    - apply (ord_trans _ _ _ (@app_eta a b)).
      now apply arrow_mon_r.
    - apply meet_elim.
      assumption.
  Qed.


  (* * We now define the abstraction of a function "f",
       denoted by "λ' f". *)

  Definition abs f := ⊓ (fun x => ∃ a, x = a ↦ f a).
  Notation "λ- f":= (abs f) (at level 60):ia_scope.

  (* * Properties of the abstraction *)
  
  
  Lemma abs_mon: ∀ f g, (∀a, f a ≤ g a) → (λ- f) ≤ (λ- g).
  Proof.
  intros f g H.
  auto_meet_leq.
  now apply arrow_mon_r.
  Qed.
  
  Lemma betarule: ∀ f a, (λ- f) @ a ≤ f a.
  Proof.
    intros f a.
    unfold app.
    apply meet_elim.
    unfold abs.
    apply meet_elim.
    exists a;reflexivity.
  Qed.


  Definition eta a := fun x => a @ x .
  Lemma etarule: ∀ a, a ≤ (λ- (eta a)).
  Proof.
    intros a.
    unfold eta,app,abs.
    apply meet_intro.
    intros y Hy.
    destruct Hy as [b Hb];rewrite Hb.
    rewrite arrow_meet.
    apply meet_intro.
    intros z Hz;red in Hz.
    destruct Hz as [Z [HZ Hz]].
    rewrite Hz; assumption.
  Qed.
  
  
  Definition upwards_closure (A:Ens):= ∀ a b, a≤ b → A a → A b .
  Definition app_closure (A:Ens):= ∀ a b, A a → A b → A (a@b).
  Definition mod_pon_closure (A:Ens):= ∀ a b, A a → A (a↦b) → A b.
  
  Remark mod_pon_app:∀ A:Ens, upwards_closure A → (app_closure A <-> mod_pon_closure A).
  Proof.
    intros A UpC. split;intros H a b Ha Hb.
    - apply (UpC ((a↦b)@a)).
      + now apply adjunction.
      + now apply H.
    - apply (H b);intuition.
      apply (UpC a).
      + now apply adjunction.
      + assumption.
  Qed.
  

End ImpStructure_lambda.
