Set Implicit Arguments.
Set Contextual Implicit.

Require Import Utf8.
Require Import RelationClasses.
Require Import Morphisms.
Require Import Setoid.
Require Import ImpAlg.Ens.
Require Import ImpAlg.Lattices.
Require Export ImpAlg.BotTop.
Require Export ImpAlg.ImpAlg.
Require Export ImpAlg.ImpAlg_Lambda.
Require Export ImpAlg.ImpAlg_Separator.
Require Export ImpAlg.ImpAlg_Adequacy.


Definition par_set_r {X:Set} {par:X → X → X} (a:X) (B:Ens):Ens :=
    fun x => ∃b, B b ∧ x = par a b.
Definition par_set_l {X:Set} {par:X → X → X} (A:Ens) (b:X):Ens :=
    fun x => ∃a, A a ∧ x = par a b.
Definition par_set {X:Set} {par:X → X → X} (A:Ens) (B:Ens):Ens :=
    fun x => ∃a,∃b, A a ∧ B b ∧ x = par a b.


Global Instance par_set_r_Proper `{X:Set} `{par:X → X → X} a:
  Proper ((@same_set X) ==> (@same_set X)) ((@par_set_r X par) a).
Proof.
  intros A B HAB.
  split;intros Hx; destruct Hx as [b [Hb Hx]];
  exists b;
  split;[apply HAB,Hb| assumption|apply HAB,Hb| assumption].
Qed.

Global Instance par_set_l_Proper `{X:Set} `{par:X → X → X} a:
  Proper ((@same_set X) ==> (@same_set X)) (fun A => (@par_set_l X par) A a).
Proof.
  intros A B HAB.
  split;intros Hx; destruct Hx as [b [Hb Hx]];
  exists b;
  split;[apply HAB,Hb| assumption|apply HAB,Hb| assumption].
Qed.


Class ParStructure `{CL:CompleteLattice} :=
  {
    par:X → X → X;
    neg: X → X;

    par_mon: ∀ a a' b b', a ≤ a' → b ≤ b' ->  par a  b ≤ par a' b';

    neg_mon: ∀ a a', ord a a' ->  ord (neg a') (neg a);

    par_meet_l: ∀ A b, par (meet_set A) b = meet_set (@par_set_l X par A b);
    par_meet_r: ∀ a B, par a (meet_set B) = meet_set (@par_set_r X par a B);

    neg_meet: ∀ A, neg (meet_set A) = join_set (image_set neg A)
  }.

Generalizable Variables X ord join meet neg par L meet_set join_set.
Hint Resolve meet_elim  meet_intro join_elim join_intro par_mon neg_mon.
Hint Resolve par_meet_l par_meet_r neg_meet.


Infix "⅋":= par (at level 76, right associativity).
Notation "¬ a":=(neg a).
Notation "a ⅋↦ b":= (¬ a ⅋ b) (at level 60, right associativity).
Definition parrow  `{PS:ParStructure} a b:= (a ⅋↦ b).

Hint Unfold parrow.


(*********************************)
(*       Separator axioms                   *)
(*********************************)

Section About_ParStructure.
  Context `{PS:ParStructure}.
  
  (** ** Properties of the ⅋ operator *)

  Lemma par_mon_l: ∀ a a' b, ord a a' ->  ord (par a b ) (par a' b ).
  Proof.
  intros a a' b H.
  apply par_mon;intuition.
  Qed.
  
  Lemma par_mon_r: ∀ a b b', b ≤ b' ->  par a  b ≤ par a b'.
  Proof.
  intros a b b' H.
  apply par_mon;intuition.
  Qed.
  
  Lemma par_top_l: ∀ a:X, (⊤ ⅋ a) = ⊤. 
  Proof.
    intros a.
    rewrite <- meet_emptyset.
    rewrite par_meet_l.
    apply meet_same_set.
    intros x;split;[intros (y,(H,_))| intros H];contradict H.
  Qed.
  
  Lemma par_top_r: ∀ a:X, (a⅋⊤) = ⊤. 
  Proof.
    intros a.
    rewrite <- meet_emptyset.
    rewrite par_meet_r.
    apply meet_same_set.
    intros x;split;[intros (y,(H,_))| intros H];contradict H.
  Qed.
    

  (** ** Properties of the ¬ operator *)
  
  Lemma neg_top: (¬⊤) = ⊥. 
  Proof.
    rewrite <- meet_emptyset.
    rewrite neg_meet.
    rewrite <- join_emptyset.
    apply join_same_set.
    intros x;split;[intros (y,(H,_))| intros H];contradict H.
  Qed.
  
  
  (** ** Properties of the arrow *)
  
  Lemma parrow_mon_l : ∀ a a' b : X, a ≤ a' → a' ⅋↦ b ≤ a ⅋↦ b.
  Proof.
    intros.
    apply par_mon_l.
    apply neg_mon.
    intuition.
  Qed.
  
  Lemma parrow_mon_r : ∀ a b b' : X, b ≤ b' → a ⅋↦ b ≤ a ⅋↦ b'.
  Proof.
    intros.
    apply par_mon_r.
    intuition.
  Qed.

  Lemma parrow_mon: ∀ a a' b b' , a' ≤ a -> b ≤ b' -> parrow a b ≤ parrow a' b'.
  Proof.
    intros.
    apply (@ord_trans _ _ _ (a' ⅋↦ b)).
    - apply parrow_mon_l; auto.
    - apply parrow_mon_r; auto.
  Qed.

  Lemma parrow_meet: ∀ (a : X) (B : Ens),  a ⅋↦ (meet_set B) = meet_set (@arrow_set_r X parrow a B).
  Proof.
    intros.
    apply par_meet_r.
  Qed.

  Definition PS_to_IS:=  @Build_ImpStructure X Ord L CL parrow parrow_mon_l parrow_mon_r parrow_meet.

  Lemma inf_neg: ∀ (f:X→X), ⋂[a] (f a) ≤ ⋂[a] (f (¬a)).
  Proof.
  intro f.
  auto_meet_leq.
  Qed.


  (** The following definition is equivalent to:
        (*   Definition app:=(@app _ _ _ _ PS_to_IS). *)
       The same apply to "abs" and "eta".
  *)

  Definition app a b:= ⊓ (fun x => a ≤ b ⅋↦ x).
  Infix "@" := app (at level 59, left associativity). 
  Definition abs f := ⋂[a] (a ⅋↦ f a).
  Notation "λ- f":= (abs f) (at level 60).
  Definition eta (a:X):=fun x => a @ x .
  
  (**  We can transfer all the properties of application 
  and abstraction obtained in implicative structures.
  We only recall the main one here.*)
  
  Lemma app_mon:∀ a a' b b', a≤ a' → b ≤ b' → a @ b ≤ a' @ b'.
  Proof.
      apply (@app_mon _ _ _ _ PS_to_IS).  
  Qed.
  
  Lemma abs_mon: ∀ f g, (∀a, f a ≤ g a) → (λ- f) ≤ (λ- g).
  Proof.
    apply (@abs_mon _ _ _ _ PS_to_IS).
  Qed.
  
  Lemma betarule: ∀ f a, (λ- f) @ a ≤ f a.
  Proof.
    apply (@betarule _ _ _ _ PS_to_IS).
  Qed.

  Lemma etarule: ∀ a, a ≤ (λ- (eta a)).
  Proof.
      apply (@etarule _ _ _ _ PS_to_IS).
  Qed.

  Lemma minimum: ∀ a b, min (fun x => a ≤ b ⅋↦ x) (a@b).
  Proof.
      apply (@minimum _ _ _ _ PS_to_IS).
  Qed.
 
  Lemma adjunction: ∀ a b c, a @ b ≤ c ↔ a ≤ b ⅋↦ c.
  Proof.
    apply (@adjunction _ _ _ _ PS_to_IS).
  Qed.
  

  Definition upwards_closure (A:Ens):= ∀ a b, a≤ b → A a → A b .
  Definition app_closure (A:Ens):= ∀ a b, A a → A b → A (a@b).
  Definition mod_pon_closure (A:Ens):= ∀ a b, A a → A (a⅋↦b) → A b.
  Remark mod_pon_app:∀ A:Ens, upwards_closure A → (app_closure A <-> mod_pon_closure A).
  Proof.
    apply (@mod_pon_app _ _ _ _ PS_to_IS).
  Qed.  

  
End About_ParStructure.

Hint Resolve par_mon_l par_mon_r parrow_mon_l parrow_mon_r parrow_mon.


(** These are the combinators as defined by Bourbaki in
    Eléments de mathématique. Théorie des ensembles. I§3, p25 *)
  


Definition PS1 `{PS:ParStructure}:= ⋂[ a ] ((a ⅋ a) ⅋↦ a).
(*  meet_set (fun x => ∃ a, x= (a ⅋ a) ⅋↦ a).*)

Definition PS2 `{PS:ParStructure}:= ⋂[ a ] (⋂[ b ] (a ⅋↦ (a ⅋ b))). 
(*   meet_set (fun x => ∃ a, x= meet_set (fun y => ∃ b, y =  a ⅋↦ (a ⅋ b))).*)

Definition PS3 `{PS:ParStructure}:= ⋂[ a ] (⋂[ b ] ((a ⅋ b)  ⅋↦  (b ⅋ a))).
(*  meet_set (fun x => ∃ a, x= meet_set (fun y => ∃ b, y = a ⅋ b ⅋↦ (b ⅋ a))).*)

Definition PS4 `{PS:ParStructure}:= ⋂[ a ] ⋂[ b ] ⋂[ c ] ((a ⅋↦ b) ⅋↦ (c ⅋ a) ⅋↦ (c ⅋ b)).
(*  meet_set (fun x => ∃ a, x= meet_set (fun y => ∃ b, y = meet_set (fun z => ∃ c, z= (a ⅋↦ b) ⅋↦ (c ⅋ a) ↦(c ⅋ b)))).*)

Definition PS5 `{PS:ParStructure}:= ⋂[ a ] ⋂[ b ] ⋂[ c ] ((a ⅋ (b ⅋ c)) ⅋↦ ((a ⅋ b) ⅋ c)).


Definition PK `{PS:ParStructure}:=
    meet_set (fun x => ∃ a, x= meet_set (fun y => ∃ b, y = a ⅋↦ b ⅋↦ a)).
Definition PS `{PS:ParStructure}:=
  meet_set (fun x => ∃ a, x= meet_set (fun y => ∃ b, y = meet_set (fun z => ∃ c, z= (a ⅋↦ b ⅋↦ c) ⅋↦ (a ⅋↦ b) ⅋↦ a ⅋↦ c))).
Definition Pcc `{PS:ParStructure} :=
  meet_set (fun x => ∃ a, x= meet_set (fun y => ∃ b, y = ((a ⅋↦ b) ⅋↦ a) ⅋↦ a)).


Class ParAlg `{PS:ParStructure}:=
  {
    separator: X → Prop;
    modus_ponens: ∀ a b, separator a → separator (¬ a ⅋ b) → separator b;
    up_closed: ∀ a b, ord a b → separator a → separator b;
    sep_PS1 : separator PS1;
    sep_PS2 : separator PS2;
    sep_PS3 : separator PS3;
    sep_PS4 : separator PS4;
    sep_PS5 : separator PS5;
  }.
Notation "a '∈ʆ'":= (separator a) (at level 80,no associativity).
Hint Resolve sep_PS1 sep_PS2 sep_PS3 sep_PS4 sep_PS5. 


(* * Tactics  *)

Ltac step ax:= eapply up_closed;[|apply ax]; repeat auto_meet_intro;  auto_meet_elim_trans. 
Ltac auto_empty_meet_elim :=  apply meet_elim_trans ;later;split;[later;[assumption|reflexivity]|auto_meet_leq].
Ltac auto_add_inf_X X:=
match goal with
  |[ |- separator ?p ]=> 
  apply up_closed with (meet_set (fun x => ∃(b:X), x=(p))) ;auto_meet_leq;[repeat auto_empty_meet_elim|]
end.



Section About_ParAlg.
  Context `{PA:ParAlg}.

  Infix "@" := app (at level 59, left associativity). 
  Notation "λ- f":= (abs f) (at level 60).

  Ltac auto_add_inf:=auto_add_inf_X X.

  Lemma modus_ponens_arrow: ∀ a b, separator a → separator (a ⅋↦ b) → separator b.
  Proof.
    intros a b Sa Sab.
    apply (modus_ponens a b Sa).
    intuition.  
  Qed.

  Definition IS:=PS_to_IS.

  Definition entails a b := separator (a ⅋↦ b).
  Infix "⊢":= entails (at level 75).
  Definition equivalence a b := a ⊢ b ∧  b ⊢ a.


  Lemma S1_p : ∀ (a:X), (a ⅋ a) ⊢ a.
  Proof.
    intros.
    apply (up_closed PS1);auto.
    auto_meet_elim_trans.
  Qed.


  Lemma S2_p : ∀ a b, a ⊢ (a ⅋ b).
  Proof.
    intros.
    apply (up_closed PS2);auto.
    auto_meet_elim_trans.
  Qed.

  Lemma S3_p : ∀ a b, (a ⅋ b)  ⊢ (b ⅋ a).
  Proof.
    intros.
    apply (up_closed PS3);auto.
    auto_meet_elim_trans.
  Qed.

  Lemma S4_p : ∀ a b c,(a ⅋↦ b) ⊢ (c ⅋ a) ⅋↦ (c ⅋ b).
  Proof.
    intros.
    apply (up_closed PS4);auto.
    auto_meet_elim_trans.
  Qed.

  Hint Resolve S1_p S2_p S3_p S4_p.



  Lemma app_closed : ∀ a b, separator a → separator b → separator (a@b).
  Proof.
  intros.
  apply (modus_ponens b (a@b)).
  - assumption.
  - apply (up_closed a).
   + apply adjunction.
     apply ord_refl.
   + assumption.
  Qed.

  Hint Resolve app_closed adjunction etarule betarule.



  Lemma C6_p a b c: (a ⊢ b) → (b ⊢ c) → (a ⊢ c).
  Proof.
    intros Sab Sbc.
    apply (modus_ponens (a ⅋↦ b) _ Sab).
    apply (modus_ponens (b ⅋↦ c) _ Sbc).
    apply S4_p.
  Qed.

(*   Fixpoint nary n X:=
    match n with
    | 0 => X
    | Datatypes.S m => X → (nary m X)
    end.
    
  Fixpoint nary_inf n (f:nary n X):=
    match n,f with
    | 0,f =>  f
    | Datatypes.S m,f => ((⋂[a] (nary_inf m (f a))))
    end .

  Fixpoint nary_arr n (f g:nary n X): nary n X:=
    match n,f,g with
    | 0,f,g => f ⅋↦ g
    | Datatypes.S m,f,g => fun a => (nary_arr m (f a) (g a))
    end.

  Fixpoint nary_par n (f g:nary n X): nary n X:=
    match n,f,g with
    | 0,f,g => f ⅋ g
    | Datatypes.S m,f,g => fun a => (nary_arr m (f a) (g a))
    end.
 *)

  Definition PC6 (f g:X→X→X):=(⋂[a] ⋂[b]((f a b) ⅋↦ (g a b))).
  Definition PC6_ter (f g:X→X→X→X):=(⋂[a] ⋂[b] ⋂[c] ((f a b c) ⅋↦ (g a b c))).
(*   Definition PC6_n n (f g:nary n X):= (nary_inf n (nary_arr n f g)). *)


  Lemma mod_pon_inf: 
     ∀ (f g:X→X→X), separator (⋂[a] ⋂[b] ((f a b) ⅋↦ (g a b))) → separator (⋂[a] ⋂[b] (f a b)) → separator (⋂[a] ⋂[b] (g a b)).
  Proof.
    intros f g Hfg Hf.
    remember (⋂[a] ⋂[b] ((f a b) ⅋↦ (g a b))) as Ifg.
    remember (⋂[a] ⋂[b] (f a b)) as If.
    apply up_closed with (Ifg@If);auto.
    repeat auto_meet_intro.
    apply adjunction.
    rewrite HeqIfg.
    auto_meet_elim_trans.
    apply parrow_mon_l.
    rewrite HeqIf.
    auto_meet_elim_trans.
  Qed.

  Lemma mod_pon_inf_ter: 
     ∀ (f g:X→X→X→X), separator (⋂[a] ⋂[b] ⋂[c] ((f a b c) ⅋↦ (g a b c)))
     → separator (⋂[a] ⋂[b] ⋂[c] (f a b c)) → separator (⋂[a] ⋂[b] ⋂[c] (g a b c)).
  Proof.
    intros f g Hfg Hf.
    remember (⋂[a] ⋂[b] ⋂[c] ((f a b c) ⅋↦ (g a b c))) as Ifg.
    remember (⋂[a] ⋂[b] ⋂[c](f a b c)) as If.
    apply up_closed with (Ifg@If);auto.
    repeat auto_meet_intro.
    apply adjunction.
    rewrite HeqIfg.
    auto_meet_elim_trans.
    apply parrow_mon_l.
    rewrite HeqIf.
    auto_meet_elim_trans.
  Qed.

 (*  Lemma mod_pon_inf_nary: 
     ∀ n (f g:nary n X), separator (PC6_n n f g) → separator (nary_inf n f) → separator (nary_inf n g).
  Proof.
    intros n f g Hfg Hf.
    apply up_closed with ((PC6_n n f g)@(nary_inf n f));auto.
    induction n;simpl.
    - apply adjunction. reflexivity.
    - cbn.
    repeat auto_meet_intro.
    apply adjunction.
    auto_meet_elim_trans.
    (*unshelve*) eapply ord_trans.
      + apply adjunction.
        instantiate (3:=a).
        apply (IHn (f a) (g a)).
        * eapply up_closed;[cbn|exact Hfg].
          auto_meet_elim_trans.
          instantiate (1:=a); reflexivity.
        * eapply up_closed;[cbn|exact Hf].
          auto_meet_elim_trans.
      + apply parrow_mon_l.
        auto_meet_elim_trans.
  Qed. *)



  Lemma sep_PC6: forall f g h, separator (PC6 f g) → separator (PC6 g h) → separator (PC6 f h).
  Proof.
    intros f g h Hfg Hgh.
    apply up_closed with ((PS4 @(PC6 g h))@(PC6 f g));auto.
    repeat auto_meet_intro.
    do 2 apply adjunction.
    auto_meet_elim_trans.
    apply parrow_mon;[auto_meet_elim_trans|].
    apply parrow_mon;[auto_meet_elim_trans|]. 
    reflexivity. 
  Qed.

  Lemma sep_PC6_ter: ∀ f g h, separator (PC6_ter f g) → separator (PC6_ter g h) → separator (PC6_ter f h).
  Proof.
    intros f g h Hfg Hgh.
    apply up_closed with ((PS4 @(PC6_ter g h))@(PC6_ter f g));auto.
    repeat auto_meet_intro.
    do 2 apply adjunction.
    auto_meet_elim_trans.
    apply parrow_mon;[auto_meet_elim_trans|].
    apply parrow_mon;[auto_meet_elim_trans|].
    reflexivity. 
  Qed.




  Definition PC7:= ⋂[a] ⋂[b] (b ⅋↦ (a ⅋ b)).
  Lemma sep_C7: separator PC7.
  Proof.
    apply sep_PC6 with (g:=fun a b => b ⅋ a).
    - eapply up_closed.
      apply inf_comm_aux.
      apply sep_PS2.
    - unfold PC6.
      eapply up_closed.
      apply inf_comm_aux.
      apply sep_PS3.
  Qed.

  Hint Resolve sep_C7.

  Lemma C7: ∀ a b, separator (b ⅋↦ (a ⅋ b)).
  Proof.
    intros a b.
    apply (up_closed PC7);[auto_meet_elim_trans|exact sep_C7].
  Qed.


  Ltac auto_step_ter p ax:=  eapply sep_PC6_ter with (g:= p);[step ax|].

  Definition par_assoc_l:=sep_PS5.
  Lemma par_assoc_r: separator (⋂[a] (⋂[b] (⋂[c] (((a ⅋ b) ⅋ c) ⅋↦ (a ⅋ (b ⅋ c)))))).
  Proof.
    auto_step_ter (fun a b c => c ⅋ (a ⅋ b)) sep_PS3.
    auto_step_ter (fun a b c => (c ⅋ a) ⅋ b) sep_PS5.  
    auto_step_ter (fun a b c => b ⅋ (c ⅋ a)) sep_PS3.  
    auto_step_ter (fun a b c => (b ⅋ c) ⅋ a) sep_PS5.  
    step sep_PS3.
  Qed.



  Definition Pswap_r:=(⋂[a] (⋂[b] (⋂[c] ((a ⅋ ( b ⅋ c)) ⅋↦ (a ⅋ (c ⅋ b)))))).
  Lemma swap_r: separator Pswap_r.
  Proof.
    eapply mod_pon_inf_ter with (f:= fun a b c => (b ⅋ c) ⅋↦ (c ⅋ b)).
    step sep_PS4. 
    step sep_PS3.
  Qed.

  Definition Pswap_l:=⋂[a] (⋂[b] (⋂[c] ((a ⅋ ( b ⅋ c)) ⅋↦ (b ⅋ (a ⅋ c))))).

  Lemma swap_l: separator Pswap_l.
  Proof.
    auto_step_ter (fun a b c => (b ⅋ c) ⅋ a) sep_PS3.  
    auto_step_ter (fun a b c => b ⅋ (c ⅋ a)) par_assoc_r.
    step swap_r.  
  Qed.


  (******** Ne sert à rien avec les axiomes en double-inf...*)
  Ltac auto_mod_pon_ter ax:=
    match ax with
    | separator (⋂[a] (⋂[b] (⋂[c] ( ?p a b c)))) => 
        try(eapply mod_pon_inf_ter with (f:=fun a b c => p a b c);[try(step ax)|try assumption]);
        try(eapply mod_pon_inf_ter with (g:=fun a b c => p a b c);[try assumption|try(step ax)])
    end.


  Ltac step_mod_pon_ter g ax := eapply mod_pon_inf_ter with (f:=g);[step ax|try assumption].

  Lemma par_comm_sep:∀ (f g :X→X→X→X), separator (⋂[a] ⋂[b] ⋂[c] ((f a b c) ⅋ (g a b c)))
    <-> separator (⋂[a] ⋂[b] ⋂[c] ((g a b c) ⅋  (f a b c))).
  Proof.
  intros f g .
  split;intro H.
  - step_mod_pon_ter (fun a b c => (f a b c) ⅋  (g a b c)) sep_PS3. 
  - step_mod_pon_ter (fun a b c => (g a b c) ⅋  (f a b c)) sep_PS3. 
  Qed.


  Lemma par_assoc_sep:∀ (f g h:X→X→X→X), separator (⋂[a] (⋂[b] (⋂[c] (((f a b c) ⅋ ( (g a b c) ⅋ (h a b c))))))) 
    <-> separator (⋂[a] (⋂[b] (⋂[c] ((((f a b c) ⅋  (g a b c)) ⅋ (h a b c)))))).
  Proof.
  intros f g h.
  split;intro H.
  - step_mod_pon_ter (fun a b c => (f a b c) ⅋  (g a b c) ⅋ (h a b c)) par_assoc_l.
  - step_mod_pon_ter (fun a b c => ((f a b c) ⅋  (g a b c)) ⅋ (h a b c)) par_assoc_r.
  Qed.

  Lemma par_swap_sep:∀ (f g h:X→X→X→X), separator (⋂[a] (⋂[b] (⋂[c] (((f a b c) ⅋ ( (g a b c) ⅋ (h a b c))))))) 
    <-> separator (⋂[a] (⋂[b] (⋂[c] ((g a b c) ⅋  (f a b c) ⅋  (h a b c))))).
  Proof.
  intros f g h.
  split;intro H.
  - step_mod_pon_ter (fun a b c => (f a b c) ⅋  (g a b c) ⅋ (h a b c)) swap_l. 
  - step_mod_pon_ter (fun a b c => (g a b c) ⅋  (f a b c) ⅋ (h a b c)) swap_l.
  Qed.

  Lemma arrow_swap_sep:∀ (f g h:X→X→X→X), separator (⋂[a] (⋂[b] (⋂[c] (((f a b c) ⅋↦ (g a b c) ⅋↦ (h a b c))))))
    <-> separator (⋂[a] (⋂[b] (⋂[c] (((g a b c) ⅋↦ (f a b c) ⅋↦ (h a b c)))))).
  Proof.
  intros f g h.
  rewrite par_swap_sep.
  intuition.
  Qed.


   
  Hint Resolve swap_r swap_l par_assoc_r.

  Lemma S6_gen:separator (⋂[a] ⋂[b] ⋂[c] ((a ⅋↦ b) ⅋↦ (b ⅋↦ c) ⅋↦ (a ⅋↦ c))).
  Proof.
    apply up_closed with (Pswap_l @ PS4);[|apply app_closed;auto].
    repeat auto_meet_intro.
    apply adjunction.
    auto_meet_elim_trans;auto.
    apply parrow_mon.
    - auto_meet_elim_trans.
    - reflexivity.
  Qed.


  Lemma pre_comp_arr_bi: ∀ f g h k, separator ( ⋂[a] ⋂[b] ⋂[c] ((f a b c) ⅋↦ (g a b c) ⅋↦ (h a b c))) → 
      separator (PC6_ter h k ) → separator ( ⋂[a] ⋂[b] ⋂[c] ((f a b c) ⅋↦ (g a b c) ⅋↦ (k a b c))).
  Proof.
    intros f g h k Ifg Ih.
    rewrite par_assoc_sep in Ifg.
    rewrite par_assoc_sep.
    apply up_closed with ((PS4 @(PC6_ter h k))@(⋂[a] ⋂[b] ⋂[c] (((¬(f a b c)) ⅋ ¬(g a b c)) ⅋ (h a b c))));auto.
    repeat auto_meet_intro.
    do 2 apply adjunction;auto_meet_elim_trans.
    apply parrow_mon;[auto_meet_elim_trans|].
    apply parrow_mon_l; auto_meet_elim_trans.
  Qed.



  Hint Unfold PC6 PC6_ter.




  Lemma swap_imp: ∀ (f g:X→X→X→X), separator (⋂[a] ⋂[b] ⋂[c] ((c⅋ (f a b c)) ⅋↦ (c ⅋ (g a b c)))) 
      → separator (⋂[a] ⋂[b] ⋂[c] (((f a b c)⅋ c) ⅋↦ ((g a b c) ⅋c))).
  Proof.
    intros f g H.
    auto_step_ter (fun a b c => (c ⅋ (f a b c))) sep_PS3.  
    auto_step_ter (fun a b c => (c ⅋ (g a b c))) H.
    step sep_PS3.
  Qed.

  Lemma mod_pon_cong: 
     ∀ (f g:X→X→X), separator (⋂[a] ⋂[b] ⋂[c] (((f a b) ⅋↦ (g a b))⅋c)) 
     → separator (⋂[a] ⋂[b] ⋂[c] ((f a b)⅋c)) → separator (⋂[a] ⋂[b] ⋂[c] ((g a b)⅋c)).
  Proof.
    intros f g Hfg Hf.
    eapply mod_pon_inf_ter ; [|exact Hf].
    simpl.
    eapply sep_PC6_ter with (g:= (fun a b c => ((g a b) ⅋ c) ⅋ c)).
    - apply swap_imp.
      step_mod_pon_ter (fun a b c => ((f a b) ⅋↦ ( (g a b) ⅋ c))) sep_PS4.
      step_mod_pon_ter (fun a b c => (((f a b) ⅋↦ (g a b)) ⅋ c)) par_assoc_r.
    - auto_step_ter (fun a b c => ((g a b) ⅋ (c ⅋ c))) par_assoc_r.
      step_mod_pon_ter (fun (a b c:X) => ((c ⅋ c) ⅋↦ (c))) sep_PS4.
      step sep_PS1.
  Qed.

  Definition PI:=(⋂[a] (a ⅋↦ a)).
  Lemma sep_PI : separator  PI.
  Proof.
    apply (up_closed (⊓ (fun x =>  ∃(b:X), x= PI))).
    -  auto_meet_intro. eapply meet_elim_trans. 
      later. split;[later;[auto|reflexivity]| auto_meet_elim_trans].
    - eapply sep_PC6 with (g:= (fun b a => a  ⅋ a)).
     + step sep_PS2.
     + step sep_PS1.
  Qed.


  Definition PS2_l := ⋂[ a ] ⋂[ b ] (a ⅋↦ (b ⅋ a)).
  Lemma sep_PS2_l: separator PS2_l.
  Proof.
  eapply sep_PC6 with (g:= (fun a b=> a  ⅋ b)).
  step sep_PS2.
  step sep_PS3.
  Qed.


  Definition PS4_r := ⋂[ a ] ⋂[ b ] ⋂[ c ] ((a ⅋↦ b) ⅋↦ (a ⅋ c) ⅋↦ (b ⅋ c)).
  Lemma sep_PS4_r: separator PS4_r.
  Proof.
    auto_step_ter (fun a b c => (c ⅋ a) ⅋↦ (c ⅋ b)) sep_PS4.
    eapply sep_PC6_ter with (g:=fun a b c => (c ⅋ a) ⅋↦ (b ⅋ c)).
    - step_mod_pon_ter (fun (a b c:X) => (c ⅋ b) ⅋↦ (b ⅋ c)) sep_PS4.
      step sep_PS3.
    - unfold PC6_ter.
      rewrite par_swap_sep.
      auto_step_ter (fun (a b c:X) =>  (c ⅋ a)) sep_PS3.
      unfold PC6_ter;  rewrite <- par_swap_sep.
      apply (up_closed PI);[|  apply sep_PI].
      auto_meet_leq. 
  Qed.



  Lemma sep_K : separator PK.
  Proof. 
    apply (up_closed PC7);auto.
    unfold PC7,PK.
    eapply ord_trans.
    - apply inf_comm_aux.
    - cbn. auto_meet_leq.
  Qed.



  Lemma arrow_precomp_par_assoc:forall (f g h k:X→X→X→X), separator (⋂[a] (⋂[b] (⋂[c] (((f a b c ) ⅋ (g a b c) ⅋ (h a b c)) ⅋↦ (k a b c))))) 
  <->     separator (⋂[a] (⋂[b] (⋂[c] ((((f a b c ) ⅋ (g a b c)) ⅋ (h a b c)) ⅋↦ (k a b c))))).
  Proof.
  intros f g h k;split;intro H.
  - auto_step_ter (fun a b c => ((f a b c ) ⅋ (g a b c) ⅋ (h a b c))) par_assoc_r;assumption. 
  - auto_step_ter (fun a b c => (((f a b c ) ⅋ (g a b c)) ⅋ (h a b c))) sep_PS5;assumption. 
  Qed.

  Lemma sep_S : separator PS.
  Proof.
    auto_step_ter (fun a b c => b ⅋↦ a ⅋↦ c) swap_l.
    auto_step_ter (fun a b c => ((a ⅋↦ b) ⅋↦ a ⅋↦ a ⅋↦ c)) sep_PS4.
    auto_step_ter (fun a b c => (a ⅋↦ a ⅋↦ c)⅋ (¬(a ⅋↦ b))) sep_PS3.
    eapply sep_PC6_ter with (g:=fun a b c => (a ⅋↦ c)⅋ (¬(a ⅋↦ b)));[|step sep_PS3].
    unfold PC6_ter.
    step_mod_pon_ter (fun (a b c:X) => (a ⅋↦ a ⅋↦ c) ⅋↦ ( a ⅋↦ c)) sep_PS4_r.
    auto_step_ter (fun (a b c:X) => ((¬ a ) ⅋ (¬ a)) ⅋ c) sep_PS5.  
    step_mod_pon_ter (fun (a b c:X) => ((¬ a) ⅋ (¬a)) ⅋↦ ¬a) sep_PS4_r.
    step sep_PS1.
  Qed.  



  Lemma ex_falso: separator (⋂[a] ( bot ⅋↦ a )).
  Proof.
    apply (up_closed PI);[|apply sep_PI].
    auto_meet_leq.
    apply parrow_mon_r.
    auto.
  Qed.

  Lemma dni: separator (⋂[a] (a ⅋↦ ¬¬a)).
  Proof.
    do 2 auto_add_inf.
    rewrite <- par_comm_sep.
    apply up_closed with (⋂[a](a ⅋↦ a));[auto_meet_leq|exact sep_PI].
  Qed.

  Lemma dni_entails: forall a, a ⊢ ¬¬a.
  Proof.
  intro a.
  eapply up_closed;[|exact dni];auto_meet_elim_trans.
  Qed.


  Ltac auto_rewrite:=
    match goal with
    |[ H: ?x = _ |- ?x = _] => rewrite H; try (reflexivity)
    |[ H: ?x = _ |- ?f ?x = _] => rewrite H; try (reflexivity)
    end.
    
  Ltac auto_same_set_step:=
    match goal with
    |[ |- _ ≅ _ ] => let x:=fresh "x" in let a := fresh "a" in let H:=fresh "H" in 
                intro x; unfold image_set,par_set_r; split;intros (a,H)
    |[ H: (∃ _: X, _ = _ ) ∧ ?x = _ |- ∃ _:X, ?x = _] => 
      let b:=fresh "a" in let Ha:=fresh "H" in let Hb:=fresh "H" in
      destruct H as ((b,Hb),Ha);try(later; rewrite Ha,Hb)
    |[ H: ?x = _ |- ∃ _ : X, ((∃ _ : X, _ = _) ∧ ?x = _ )] => 
        try(later;split;[later;reflexivity|]);try(rewrite H;now reflexivity)
    |[ H: ?x = _ |- ∃ _ : X, ?x = _] => later;repeat(auto_rewrite)
    |[ |- ⊓ _ = ⊓ _] => apply meet_same_set
    |[ |-  _ =  _] => now reflexivity
    end.
  Ltac auto_same_set:=repeat (auto_same_set_step).


  Lemma par_meet_f_r: ∀ (b:X) (f:X→X),( b ⅋ (⋂[a] f a)) = ⋂[a] (b ⅋ ( f a)).
  Proof.
  intros b f.
  rewrite par_meet_r.
  apply meet_same_set.
  auto_same_set.
  Qed.

  Lemma arrow_meet_r: ∀ (b:X) (f:X→X),( b ⅋↦ (⋂[a] f a)) = ⋂[a] (b ⅋↦ ( f a)).
  Proof.
  intros b f.
  apply par_meet_f_r.
  Qed.

  Lemma arrow_meet_r_aux: ∀ (f:X→X→X) (g:X→X→X→X),
  ⋂[a] (⋂[b] ((f a b) ⅋↦ (⋂[c] (g a b c)))) = ⋂[a] ⋂[b] ⋂[c] ((f a b ) ⅋↦  (g a b c)).
  Proof.
  intros f g.
  apply ord_antisym.
  - auto_meet_leq. apply parrow_mon_r. auto_meet_leq.
  - do 2 auto_meet_intro. apply adjunction. auto_meet_intro. apply adjunction. auto_meet_elim_trans.
  Qed.


  Lemma extra_hyp: forall (f g h:X→X→X), separator (⋂[a] ⋂[b] ((f a b ) ⅋↦ (g a b))) →
  separator (⋂[a] ⋂[b] ((h a b) ⅋↦ (f a b) ⅋↦ (g a b ))).
  Proof.
  intros f g h Hfg.
  eapply mod_pon_inf with (fun a b => (f a b ) ⅋↦ (g a b));auto.
  step sep_PS2_l.
  Qed.

  Lemma extra_hyp_ter: forall (f g h:X→X→X→X), separator (⋂[a] ⋂[b] ⋂[c] ((f a b c) ⅋↦ (g a b c))) →
  separator (⋂[a] ⋂[b] ⋂[c] ((h a b c) ⅋↦ (f a b c) ⅋↦ (g a b c))).
  Proof.
  intros f g h Hfg.
  step_mod_pon_ter (fun a b c => (f a b c) ⅋↦ (g a b c)) sep_PS2_l.
  Qed.

  Lemma half_modus_ponens: ∀ (f g:X→X→X), separator (⋂[a] ⋂[b] (f a b))
  → separator (⋂[a] ⋂[b] (((f a b ) ⅋↦ (g a b)) ⅋↦ (g a b))).
  Proof.
  intros  f g Hf.
  eapply mod_pon_inf;[|exact Hf].
  auto_add_inf.
  rewrite arrow_swap_sep.
  step sep_PI.
  Qed.




  Definition or a b:= ⋂[c] ((a ⅋↦ c) ⅋↦ (b ⅋↦ c) ⅋↦ c).
  Infix "+":=or.

  Lemma par_or_gen : separator (⋂[a] ⋂[b]( (a ⅋ b) ⅋↦ or a b)).
  Proof.
  unfold or.
  rewrite arrow_meet_r_aux.
  assert (⋂[ a] (⋂[ b] (⋂[ c] ( (b ⅋↦ c) ⅋↦ (a ⅋ b) ⅋↦  (a ⅋↦ c)  ⅋↦ c))) ∈ʆ).
  - apply (pre_comp_arr_bi (fun a b c => b ⅋↦ c) (fun a b c => a ⅋ b ) 
          (fun (a b c:X) => a ⅋ c) (fun (a b c:X) => (¬ a ⅋↦ c) ⅋ c)).
    step sep_PS4.
    auto_step_ter (fun (a b c:X) => c⅋a) sep_PS3.
    unfold PC6_ter.
    rewrite arrow_swap_sep.
    apply pre_comp_arr_bi with (h:=fun (a b c:X) => c⅋c);[step sep_PS4|step sep_PS1].
  - rewrite par_assoc_sep .  
    rewrite par_swap_sep.
    auto_step_ter (fun (a b c:X) => (a ⅋ b) ⅋↦ (a ⅋↦ c) ⅋↦ c) H.
    step sep_PS5.
  Qed.

  Lemma par_or : ∀ a b, (a ⅋ b) ⊢ (or a b).
  Proof.
  intros a b. step par_or_gen.
  Qed.

  Lemma or_par_gen : separator (⋂[a] ⋂[b]( or a b ⅋↦ (a ⅋ b))).
  Proof.
  apply sep_PC6 with (g:= fun a b => ((a ⅋↦ (a⅋b))⅋↦ (b ⅋↦ (a⅋b)) ⅋↦ (a⅋b))).
  - step sep_PI.
    apply parrow_mon_r. auto_meet_elim_trans. 
  - apply sep_PC6 with (g:= fun a b => ((b ⅋↦ (a⅋b)) ⅋↦ (a⅋b))).
    + apply half_modus_ponens. step sep_PS2. 
    + apply half_modus_ponens. step sep_PS2_l. 
  Qed.

  Lemma or_par : ∀ a b, (or a b) ⊢ (a ⅋ b).
  Proof.
  intros a b. step or_par_gen.
  Qed.


  Lemma neg_arrow: separator (⋂[a] ((¬a) ⅋↦ (a ⅋↦ ⊥))).
  Proof.
  auto_add_inf.
  rewrite inf_comm.
  step sep_PS2.
  Qed.

  Lemma neg_imp_bot: ∀ a, (¬ a) ⊢ a ⅋↦ ⊥.
  Proof.
  intro a. step neg_arrow.
  Qed. 

  Lemma arrow_neg: separator (⋂[a] ((a ⅋↦ ⊥) ⅋↦ (¬a))).
  Proof.
  auto_add_inf.
  apply sep_PC6 with (g:= fun a b=> (b ⅋↦ ¬ b)).
  - step sep_PI. apply parrow_mon_r. intuition. 
  - apply mod_pon_inf with (f:=fun a b => (¬b) ⅋↦ ¬b);[|step sep_PI].
    apply mod_pon_inf with (f:=fun (a b:X) => (¬ b) ⅋ b);[|step sep_PI].
    step par_or_gen. 
    apply parrow_mon_r.
    auto_meet_elim_trans.
  Qed.

  Lemma imp_bot_neg: ∀ a, (a ⅋↦ ⊥) ⊢ ¬ a.
  Proof.
  intro a. step arrow_neg.
  Qed. 



  Lemma absurd_aux_ter: 
  ∀ (f:X→X→X→X), separator (⋂[a] (⋂[b] ⋂[c] (((¬ (f a b c)) ⅋↦ (f a b c)) ⅋↦ (f a b c)))).
  Proof.
  intros f.
  eapply mod_pon_inf_ter with (f:=fun (a b c:X)=> (f a b c) ⅋↦ (f a b c));[|step sep_PI].
  eapply mod_pon_inf_ter with (f:=fun (a b c:X)=> (f a b c)⅋ ¬(f a b c)).
  - eapply up_closed;[|exact par_or_gen].
    unfold or.
    rewrite arrow_meet_r_aux.
    auto_meet_leq.
  - step_mod_pon_ter (fun (a b c:X) => (¬ (f a b c))⅋ (f a b c)) sep_PS3.
    step sep_PI.
  Qed.

  Lemma absurd_aux: separator (⋂[a] (((¬ a) ⅋↦ a) ⅋↦ a)).
  Proof.
  eapply up_closed;[|exact (absurd_aux_ter (fun (a b c:X)=> a))].
  auto_meet_leq.
  do 2 auto_empty_meet_elim.
  Qed.


  Lemma absurd: ∀ (f:X→X→X→X), separator (⋂[a] (⋂[b] ⋂[c] (((¬ (f a b c)) ⅋↦ ⊥) ⅋↦ f a b c))).
  Proof.
  intros f.
  apply sep_PC6_ter with (g:=fun (a b c:X) => (¬ (f a b c)) ⅋↦ (f a b c));[|step absurd_aux].
  step_mod_pon_ter (fun a b c => ⊥ ⅋↦ (f a b c)) sep_PS4.
  step ex_falso.
  Qed.


  Lemma reductio_ad_absurdum:∀a, (¬ a) ⅋↦ ⊥ ⊢ a.
  Proof.
  intros a.
  step (absurd (fun a b c => c)).
  do 2 auto_empty_meet_elim.
  Qed.



  Lemma dne: separator (⋂[a] ((¬¬a) ⅋↦ a)).
  Proof.
  do 2 auto_add_inf.
  auto_step_ter (fun (a b c:X) => ((¬c) ⅋↦ ⊥) ) neg_arrow.
  step (absurd (fun a b c => c)).
  do 2 auto_empty_meet_elim.
  Qed.

  Lemma dne_entails: forall a, (¬ ¬ a) ⊢ a.
  Proof.
  intro a.
  eapply up_closed;[|exact dne];auto_meet_elim_trans.
  Qed.




  Lemma sep_cc : separator Pcc.
  Proof.
  eapply sep_PC6 with (g:=fun a b => ((¬ a) ⅋↦ a ⅋↦ b) ⅋↦ (¬ a) ⅋↦ a).
  - step sep_PS4.
  - unfold PC6. 
    apply sep_PC6 with (g:=fun (a b:X) => ((¬ a) ⅋↦ a));unfold PC6.
    + apply half_modus_ponens.
      eapply sep_PC6 with (g:=fun (a b:X) => (a ⅋↦ ⊥));unfold PC6.
      step neg_arrow.
      step sep_PI.
      apply parrow_mon;[reflexivity|];intuition.
    + step absurd_aux.
  Qed.


  Lemma ImpK : (@K _ _ _ _ IS)= PK .
  Proof.
    intuition.
  Qed.
  Lemma ImpS : (@S _ _ _ _ IS) = PS .
  Proof.
    intuition.
  Qed.
  Lemma Impcc : (@cc _ _ _ _ IS) = Pcc.
  Proof.
    intuition.
  Qed.


  Lemma sep_ImpK : separator (@K _ _ _ _ IS).
  Proof.
    rewrite ImpK.
    apply sep_K.
  Qed.
  Lemma sep_ImpS : separator (@S _ _ _ _ IS).
  Proof.
    rewrite ImpS.
    apply sep_S.
  Qed.
  Lemma sep_Impcc : separator (@cc _ _ _ _ IS).
  Proof.
    rewrite Impcc.
    apply sep_cc.
  Qed.

  Definition ParAlg_to_ImpAlg:=@Build_ImpAlg X Ord L CL IS separator modus_ponens_arrow up_closed sep_ImpK sep_ImpS sep_Impcc.

  Definition IA_sep:=@ImpAlg_Separator.separator X Ord L CL IS ParAlg_to_ImpAlg.

  Lemma psep_if_sep: ∀ a, IA_sep a → separator a.
  Proof.
    intros.
    apply H.
  Qed.
  Lemma sep_if_psep: ∀ a, IA_sep a → separator a.
  Proof.
    intros.
    apply H.
  Qed.

  Lemma psep_iff_sep: ∀ a, IA_sep a <-> separator a.
  Proof.
    intros;split;[apply psep_if_sep|apply sep_if_psep].
  Qed.
End About_ParAlg.



