Set Implicit Arguments.
Set Contextual Implicit.

Require Import Utf8.
Require Import RelationClasses.
Require Import Logic.
Require Import Notations.
Require Import ImpAlg.Ens.
Require Import ImpAlg.Lattices.
Require Export ImpAlg.BotTop.
Require Export ImpAlg.ParAlg.

Section LPar_Encoding.
  Context `{PS:ParStructure}.
  
  Definition ok c := fst c ≤ snd c.
  Global Instance ok_Reflexive : Reflexive (fun x y=> ok (x,y)):=ord_refl.
  
  Notation "c '∈(≤)'":=(ok c) (at level 60,no associativity).
  Notation "'<' t '|' u '>'":=(pair t u) (at level 60,no associativity).
  
  
  Definition cord p p':= ok p → ok p'.

  Infix "◃":=(cord) (at level 70).
  Global Instance cord_Reflexive : Reflexive cord.
  Proof.
   intros x H. exact H.
  Qed. 
  Global Instance cord_Transitive : Transitive cord.
  Proof.
   intros x y z Hx Hy H. apply (Hy (Hx H)).
  Qed. 

  
  
  Lemma cord_mon: ∀ t u t' u', t' ≤ t → u ≤ u' → (t, u) ◃ (t',u').
  Proof.
    intros t u t' u' Ht Hu H.
    apply (ord_trans _ _ _ Ht);simpl.
    apply (ord_trans _ _ _ H Hu).
  Qed.
  
  Lemma cord_subset: ∀ c c', (∀ a, c a ◃ c' a) → subset  (λ x:X, ∃a, x = a ∧ ok (c a)) (λ x, ∃a, x= a ∧ ok (c' a)).
  Proof.
    intros c c' H x (a,(Hx,Ha)).
    subst x.
    exists a;intuition.
    now apply H.
  Qed.
    
  Lemma cord_meet: ∀ c c', (∀ a, c a ◃ c' a) → ⊓(λ x, ∃ a, x=a ∧ ok (c' a)) ≤ ⊓(λ x, ∃a, x= a ∧ ok (c a)).
  Proof.
  intros c c' H.
  apply meet_subset.
  now apply cord_subset.
  Qed.


  (** * Encodings of the positives terms  of the L⅋-calculus *)
  Definition box (t:X):= ¬ t.  
  Definition pairing a b:= a⅋b.  
  Definition mup c:= ⊔(λ x, ∃ a, x= a ∧ ok (c a )).  
  Notation "'µ+.' c":= (mup c) (at level 60,no associativity).
  Notation "'{' t ',' u '}'":= (pairing t u) (at level 60,no associativity).
  Notation "'[' t ']'":= (box t) (at level 60,no associativity).
  
  
  (** ** Properties of the encoding of µ+ *)
   
  Lemma mup_mon: ∀ c c', (∀ a, (c a)◃ (c' a)) → µ+.c ≤ µ+.c'.
  Proof.
    intros c c' H. unfold mup.
    auto_join_leq.
    auto_join_elim_risky.
    now apply H.
  Qed.
  
  (** The following reflect the property of closure of the pole under anti-reduction.
      Since the ß-reduction gives:  
      
                        <t|µx.c>  →  c[t/x]
                     
      by closure under anti-reduction, if c[t/x] ∈ ⫫   then  <t|µx.c> ∈ ⫫,
      or with the notations:

                          c[t/x] ◃ <t|µx.c>
  *)                  
  
  Lemma mup_beta: ∀ c a, (c a) ◃ (a,µ+.c) .
  Proof.
    intros c a H.
    unfold mup.
    auto_join_elim_risky.
    now apply H.
  Qed.
  
  
   (** The following is the usual property of η-expansion:
                         t  →  µx.<x|t>   
      which is reflected as in implicative structures by:
                          t ≤ µx.<x|t>

  *)
  
  Lemma mup_eta: ∀ t, t ≤ µ+.(λ a, (a,t))  .
  Proof.
    intros t.
    unfold mup.
    auto_join_elim_risky.
  Qed.
  
  
  
  
  
  (** * Encodings of negative terms of the L⅋-calculus *)
  
  Definition mun c:= ⊓(λ x, ∃ a, x= a ∧ ok (c a ) ).
  Definition mu_pair c:= ⊓(λ x, ∃ a, x=⊓(λ y, ∃ b, y= (a ⅋ b) ∧ ok (c a b))).
  Definition mu_neg c:= ⊓(λ x, ∃ a, x= (¬ a) ∧ ok (c a )).
  
  Notation "'µ-.' c":= (mun c) (at level 60,no associativity).
  Notation "'µ[].' c":= (mu_neg c) (at level 60,no associativity).
  Notation "'µ{}.' c":= (mu_pair c) (at level 60,no associativity).
  
  

  (** ** Properties of the encodings of µ-.c *)
  
  Lemma mun_mon: ∀ c c', (∀ a, (c a)◃ (c' a)) → µ-.c' ≤ µ-.c.
  Proof.
    intros c c' H. unfold mun.
    auto_meet_leq.
    auto_meet_elim_risky.
    now apply H.
  Qed.
  
  
  
  
  Lemma mun_beta: ∀ c t, (c t) ◃ (µ-.c,t) .
  Proof.
    intros c t H.
    unfold mun.
    auto_meet_elim_risky.
  Qed.

 
  
  Lemma mun_eta: ∀ t, t ≤ mun (λ a,(t,a))  .
  Proof.
    intros t.
    unfold mun.
    auto_meet_leq.
    now assumption.
  Qed.

  
  (** ** Properties of the encodings of µ(x,y).c *)
  
  Lemma mu_pair_mon: ∀ c c', (∀ a b, (c a b)◃ (c' a b)) → µ{}.c' ≤ µ{}.c.
  Proof.
    intros c c' H. unfold mu_pair.
    auto_meet_leq.
    auto_meet_elim_risky.
    now apply H.
  Qed.

  Lemma mu_pair_beta: ∀ c a b, (c a b) ◃ (µ{}.c,{a,b}) .
  Proof.
    intros c a b H.
    unfold mun.
    auto_meet_elim_risky.
    now apply H.
    reflexivity.
  Qed.
  
  Lemma mu_pair_eta: ∀ t, t ≤ µ{}.(λ a b, (t,{a,b}))  .
  Proof.
    intros t.
    unfold mu_pair.
    auto_meet_leq.
    now assumption.
  Qed.

  
   (** ** Properties of the encoding of µ[a].c *)
   
  Lemma mu_neg_mon: ∀ c c', (∀ a, (c a)◃ (c' a)) → µ[].c' ≤ µ[].c.
  Proof.
    intros c c' H. unfold mu_neg.
    auto_meet_leq.
    auto_meet_elim_risky.
    now apply H.
  Qed.
  
  
  Lemma mu_neg_beta: ∀ c a, (c a) ◃ (µ[].c,[a]) .
  Proof.
    intros c a H.
    unfold mu_neg.
    auto_meet_elim_risky.
    now apply H.
    reflexivity.
  Qed.
  
  Lemma mu_neg_eta: ∀ t, t ≤ µ[].(λ a, (t,[a]))  .
  Proof.
    intros t.
    unfold mu_neg.
    auto_meet_leq.
    now assumption.
  Qed.
  

  (** ** Encodings of µ([α],x).c *)
  
  Definition mu_abs c:= ⊓(λ x, ∃ (a:X), x=⊓(λ y, ∃ b, y= ((¬a) ⅋ b) ∧ ok (c a b))).
  Notation "'µ{[],}.' c":= (mu_abs c) (at level 62,no associativity).
  
  Lemma mu_abs_char: ∀ c, µ{[],}.c = µ{}.(λ x y, (µ[].(λ z, c z y),x)).
  Proof.
    intros.
    unfold mu_abs,mu_pair.
    rewrite meet_comm.
    rewrite (meet_comm _ (fun a b => a⅋b)).
    apply meet_same_set.
    split;intros (b,Hx);exists b;subst x.
    - apply ord_antisym.
      + auto_meet_intro.
        eapply ord_trans;[|apply par_mon_l; apply H0].
        simpl.
        unfold mu_neg.
        rewrite par_meet_l.
        auto_meet_intro.
        intros y (d,((e,(He,Hf)),Hd)).
        subst d. subst y.
        auto_meet_elim_risky.
      + auto_meet_intro.
        auto_meet_elim_risky.
        * exact H0.
        * reflexivity.
    - apply ord_antisym; auto_meet_intro.
      + auto_meet_elim_risky.
        * exact H0.
        * reflexivity.
      + eapply ord_trans;[|apply par_mon_l; apply H0].
        simpl.
        unfold mu_neg.
        rewrite par_meet_l.
        auto_meet_intro.
        intros y (d,((e,(He,Hf)),Hd)).
        subst d. subst y.
        auto_meet_elim_risky.
  Qed.  
  
(** * Combinators *)
(** Combinators S1,S2,S3,S4 & S5 expressed as encodings of L⅋-terms
are equal to their principal types *)

  Lemma PS1_lpar: PS1 = µ{[],}.(λ x y,  (x,{y,y})). 
  Proof.
  apply ord_antisym.
  - auto_meet_leq.
    apply parrow_mon_l.
    apply H1.
  - auto_meet_leq.
    auto_meet_elim.
    later;split;[reflexivity|]. 
    reflexivity.
  Qed.  
  
  
  Lemma PS2_lpar_aux: ∀(x:X), µ{}.(λ w z, (x,w))= (x⅋⊥).
  Proof.
  intros x.
  apply ord_antisym.
  - auto_meet_leq. auto_meet_elim_risky. 
  - auto_meet_leq. apply par_mon ;intuition.
  Qed.
  
  Lemma PS2_lpar: PS2 = µ{[],}.(λ x y, (µ{}.(λ w z, (x,w)),y)).
  Proof.
  apply meet_same_set.
  split; intros (a,Ha);exists a;subst x;rewrite PS2_lpar_aux.
  - apply ord_antisym.
    + auto_meet_leq. apply parrow_mon_r; apply H0.
    + auto_meet_leq. auto_meet_elim_risky. apply parrow_mon_r. now apply par_mon_r.
  - apply ord_antisym.
    + auto_meet_leq. auto_meet_elim_risky. apply parrow_mon_r. now apply par_mon_r.
    + auto_meet_leq. apply parrow_mon_r; apply H0.
  Qed.  

  
  Lemma PS3_lpar: PS3 = µ{[],}.(λ x y, (µ{}.(λ w z, (x,{z,w})),y)).
  Proof.
  apply ord_antisym.
  - do 2 auto_meet_intro.
    eapply ord_trans;[|apply parrow_mon_r;apply H1].
    simpl.
    apply adjunction.
    do 2 auto_meet_intro.
    apply adjunction.
    auto_meet_elim_trans.
    apply parrow_mon;[apply H4|reflexivity].
  - do 2 auto_meet_intro.
    auto_meet_elim_risky.
    reflexivity.
  Qed.



(** Terms on the right is exactly  µ([α],x).< µ([ß],y).<µ(c,b).< ß |(c,µγ.< α | ([γ],b)>) > | y > | x > *)


(*   Lemma PS4_lpar_aux: ∀ a b, a ≤ {[µ+.(λ γ : X, ( a , {[γ], b} ))],b}.
  Proof.
    intros a b.
    auto_join_elim_risky.
    Check mup_eta.
  
  Qed.
 *)
  Lemma PS4_lpar: µ{[],}.(λ α x,
                        (µ{[],}.(λ ß y,(µ{}.(λ c b,(ß,
                                                   {c,µ+.(λ γ,(α,
                                                               {[γ],b}))}))
                                        ,y))
                        ,x)) ≤ PS4.

  Proof.
(*   apply ord_antisym.
  - do 2 auto_meet_intro.
    clear H0 H. 
    eapply ord_trans;[|apply parrow_mon_r;apply H1];simpl.
    apply adjunction.
    do 2 auto_meet_intro.
    clear H0 H. 
    eapply ord_trans;[|apply parrow_mon_r;apply H2];simpl.
    apply -> adjunction.
    do 2 auto_meet_intro.
    clear H0 H. 
    (apply <- adjunction).
    eapply ord_trans;[|apply parrow_mon_l;apply H3];simpl.
    apply adjunction.
    eapply ord_trans.
    Focus 2.
    apply parrow_mon.
    Focus 2.
    
    eapply ord_trans;[|apply parrow_mon_l;apply mup_eta];simpl.
    auto_meet_elim_trans.
    apply parrow_mon;[|apply parrow_mon_r;reflexivity];simpl.
    Focus 2.
    
    ;[|reflexivity].
    auto_meet_elim_risky. *) 
  - do 3 auto_meet_intro.
    auto_meet_elim_risky.
    + apply ord_refl. 
    + simpl. apply parrow_mon_l.
    apply par_mon_r. auto_join_elim_risky.
  Qed.


  Lemma PS5_lpar: PS5 = µ{[],}.(λ α x,(µ{}.(λ y x3 ,(µ{}.(λ x1 x2,(α,{x1,{x2,x3}})),y)),x)).
  Proof.
   apply ord_antisym.
  - do 2 auto_meet_intro.
    clear H0 H. 
    eapply ord_trans;[|apply parrow_mon_r;apply H1];simpl.
    apply adjunction.
    do 2 auto_meet_intro.
    clear H0 H. 
    eapply ord_trans;[|apply par_mon_l;apply H2];simpl.
    unfold mu_pair. rewrite par_meet_l. unfold par_set_l.
    auto_meet_intro.
    intros x3 (a3,((a4,H4),H)). subst x3 a3. 
    rewrite par_meet_l. unfold par_set_l.
    auto_meet_intro.
    intros x3 (a3,((a5,(H4,H5)),H)). subst x3 a3. 
    apply adjunction.
    auto_meet_leq.
    apply parrow_mon;[apply H5|reflexivity].
  - do 3 auto_meet_intro.
    auto_meet_elim_risky.
    + apply ord_refl. 
    + simpl. now apply par_mon_l. 
  Qed.


  
  
End LPar_Encoding.  

  