Set Implicit Arguments.
Set Contextual Implicit.

Require Import Utf8.
Require Import Setoid.
Require Import ImpAlg.Ens.
Require Import ImpAlg.Lattices.
Require Import ImpAlg.BotTop.
Require Export ImpAlg.ImpAlg.
Require Export ImpAlg.ParAlg.


(** * Tensor Structures *)
Definition tensor_set_l `{X:Set} `{tensor:X → X → X} (A:Ens) (b:X):Ens :=
    fun x => ∃a, A a ∧ x = tensor a b.
Definition tensor_set_r `{X:Set} `{tensor:X → X → X} (a:X) (B:Ens):Ens :=
    fun x => ∃b, B b ∧ x = tensor a b.
Definition tensor_set `{X:Set} `{tensor:X → X → X} (A:Ens) (B:Ens):Ens :=
    fun x => ∃a,∃b, A a ∧ B b ∧ x = tensor a b.
Notation "a ◃ b":= (rev a b) (at level 70).

Class TensorStructure `{CL:CompleteLattice} :=
  {
    tensor:X → X → X;
    neg: X → X;
    
    tensor_mon: ∀ a a' b b',  a ≤ a' -> b ≤ b' -> (tensor a b ) ≤ (tensor a' b');

    neg_mon: ∀ a a', ord a a' ->  ord (neg a') (neg a);
   
    tensor_join_l: ∀ A b, tensor (join_set A) b = join_set (@tensor_set_l X tensor A b);
    tensor_join_r: ∀ a B, tensor a (join_set B) = join_set (@tensor_set_r X tensor a B);
    
    neg_join: ∀ A, neg (join_set A) = meet_set (image_set neg A)
  }.

Generalizable Variables X ord join meet neg par L meet_set join_set.
Infix " ⨂ ":= tensor (at level 76, right associativity).
Notation "a ⨂↦ b":= (neg( a ⨂ neg b)) (at level 60, right associativity).
Definition tarrow  `{TS:TensorStructure} a b:= (a ⨂↦ b).


(** * Properties *)

Section About_TensorStruct.
Context `{TS:TensorStructure}.

 (** ** Properties of the ⨂ operator *)

  Lemma tensor_mon_l: ∀ a a' b, a ≤ a' ->  (a ⨂ b) ≤ (a' ⨂ b).
  Proof.
  intros a a' b H.
  apply tensor_mon;intuition.
  Qed.
  
  Lemma tensor_mon_r: ∀ a b b', b ≤ b' ->  (a ⨂ b) ≤ (a ⨂ b').
  Proof.
  intros a b b' H.
  apply tensor_mon;intuition.
  Qed.

  Lemma tensor_bot_l: ∀ a:X, (⊥ ⨂ a) = ⊥. 
  Proof.
    intros a.
    rewrite <- join_emptyset.
    rewrite tensor_join_l.
    apply join_same_set.
    intros x;split;[intros (y,(H,_))| intros H];contradict H.
  Qed.
  
  Lemma tensor_bot_r: ∀ a:X, (a ⨂ ⊥) = ⊥. 
  Proof.
    intros a.
    rewrite <- join_emptyset.
    rewrite tensor_join_r.
    apply join_same_set.
    intros x;split;[intros (y,(H,_))| intros H];contradict H.
  Qed.
    

  (** ** Properties of the ¬ operator *)
  
  Lemma neg_bot: (neg ⊥) = ⊤. 
  Proof.
    rewrite <- join_emptyset.
    rewrite neg_join.
    rewrite <- meet_emptyset.
    apply meet_same_set.
    intros x;split;[intros (y,(H,_))| intros H];contradict H.
  Qed.

End About_TensorStruct.




Hint Resolve tensor_mon neg_mon tensor_join_l tensor_join_r neg_join.


(** **     Reversing a ⅋-structure           *)

Section ReversePS.
  Context `{PS:ParStructure}.

  Definition preneg_p := preimage (fun a => ParAlg.neg a).
  Definition preneg_p_set:= preimage_set (fun a => ParAlg.neg a).
  Notation "¬⁻¹ a":=(preneg_p a) (at level 75).


  Definition tens a b:=par a b.
  
  Ltac autorev:=intros;unfold rev,reverse;auto.
  
 
 Lemma rev_tensor_mon: ∀ a a' b b', a ◃ a' -> b ◃ b' ->  par a  b ◃ par a'  b'.
  Proof.
    autorev.
  Qed.
  
  
  Lemma rev_neg_mon: ∀ a a', a ◃ a' -> (¬ a') ◃ (¬ a).
  Proof.
    autorev.
  Qed.

  Lemma rev_tensor_join_l: ∀ A b, par (meet_set A) b = meet_set (@tensor_set_l X par A b).
  Proof.
    unfold tensor_set_l.
    apply par_meet_l.
  Qed.
  
  Lemma rev_tensor_join_r: ∀ a B, par a (meet_set B) = meet_set (@tensor_set_r X par a B).
  Proof.
    unfold tensor_set_l.
    apply par_meet_r.
  Qed.

  Lemma rev_neg_join: ∀ A, ParAlg.neg (meet_set A) = join_set (image_set ParAlg.neg A).
  Proof.
    apply neg_meet.
  Qed.  
End ReversePS.

Definition PS_to_TS`{PS:ParStructure}:=
@Build_TensorStructure X RevOrd RevLattice RevCompleteLattice par (ParAlg.neg) 
rev_tensor_mon rev_neg_mon rev_tensor_join_l rev_tensor_join_r rev_neg_join.


Theorem Rev_ParStructure_is_TensorStructure `{CL:CompleteLattice}: 
      ParStructure → (@TensorStructure _ _ _ RevCompleteLattice ).
Proof.
intros; apply PS_to_TS.
Qed.



(** **    Reversing a ⨂-structure          *)


Section ReverseTS.
  Context `{TS:TensorStructure}.

  Notation "¬ a":=(neg a).
  Definition preneg_t := preimage (fun a => ¬ a).
  Definition preneg_t_set:= preimage_set (fun a => ¬ a).
  Notation "¬⁻¹ a":=(preneg_t a) (at level 75):ia_scope.

 
 Ltac autorev:=intros;unfold rev,reverse;auto.

 Lemma rev_par_mon: ∀ a a' b b', a ◃ a' -> b ◃ b' -> tensor a  b ◃ tensor a'  b'.
  Proof.
    autorev.
  Qed.
    
  Lemma rev_neg_p_mon: ∀ a a', a ◃ a' -> (¬ a') ◃ (¬ a).
  Proof.
    autorev.
  Qed.

  Lemma rev_par_meet_l: ∀ A b, tensor (join_set A) b = join_set (@par_set_l X tensor A b).
  Proof.
    unfold par_set_l. 
    apply tensor_join_l.
  Qed.
  
  Lemma rev_par_meet_r: ∀ a B, tensor a  (join_set B) = join_set (@par_set_r X tensor a B).
  Proof.
    unfold par_set_l. 
    apply tensor_join_r.
  Qed.

  Lemma rev_neg_meet: ∀ A, neg (join_set A) = meet_set (image_set neg A).
  Proof.
    apply neg_join.
  Qed.

End ReverseTS.


Definition TS_to_PS `{TS:TensorStructure}:=
@Build_ParStructure X RevOrd RevLattice RevCompleteLattice tensor neg 
rev_par_mon rev_neg_p_mon rev_par_meet_l rev_par_meet_r rev_neg_meet.


Theorem Rev_TensorStructure_is_ParStructure `{CL:CompleteLattice}: 
      TensorStructure → (@ParStructure _ _ _ RevCompleteLattice ).
Proof.
intros; apply TS_to_PS.
Qed.



Definition TS1 `{TS:TensorStructure}:= ⋂[ a ] (a ⨂↦ (a ⨂ a)).

Definition TS2 `{TS:TensorStructure}:= ⋂[ a ] (⋂[ b ] ((a ⨂ b) ⨂↦ a )). 

Definition TS3 `{TS:TensorStructure}:= ⋂[ a ] (⋂[ b ] ((a ⨂ b)  ⨂↦  (b ⨂ a))).

Definition TS4 `{TS:TensorStructure}:= ⋂[ a ] ⋂[ b ] ⋂[ c ] ((a ⨂↦ b) ⨂↦ (c ⨂ a) ⨂↦ (c ⨂ b)).

Definition TS5 `{TS:TensorStructure}:= ⋂[ a ] ⋂[ b ] ⋂[ c ] ((a ⨂ (b ⨂ c)) ⨂↦ ((a ⨂ b) ⨂ c)).


Class TensorAlg `{TS:TensorStructure}:=
  {
    separator: X → Prop;
    modus_ponens: ∀ a b, separator a → separator (a ⨂↦ b) → separator b;
    up_closed: ∀ a b, ord a b → separator a → separator b;
    sep_TS1 : separator TS1;
    sep_TS2 : separator TS2;
    sep_TS3 : separator TS3;
    sep_TS4 : separator TS4;
    sep_TS5 : separator TS5;
  }.
  
Notation "a '∈ʆ'":= (separator a) (at level 80,no associativity).
Hint Resolve sep_TS1 sep_TS2 sep_TS3 sep_TS4 sep_TS5. 


Section ReverseParAlg.
  Context `{PA:ParAlg}.


  Definition preneg := preimage (fun a => ParAlg.neg a).
  
  Definition preneg_set:= preimage_set (fun a => ¬ a). 
  Notation "¬⁻¹ a":=(preneg a) (at level 75):ia_scope.
  Notation "a ◃ b":= (rev Ord a b) (at level 70):ia_scope.


  
Definition TS:=PS_to_TS.
Definition psep := ParAlg.separator.
Definition tsep := preneg_set (psep).

Notation "a ⨂ b":= (@tensor _ _ _ _ TS a b).

Lemma in_preneg : ∀ A a, preneg_set A a ↔  A (¬a).
Proof.
  intros;split;intro H.
  - destruct H as (b,(Ab,Hb));subst;intuition.
  - exists (¬ a); auto.
Qed.  

Lemma in_tsep : ∀ a, tsep a ↔ psep (¬a).
Proof.
  intro a;unfold tsep.
  rewrite in_preneg.
  intuition.
Qed.

Lemma up_closed_tsep: ∀ a b, a ◃ b → tsep a → tsep b.
Proof.
intros a b Hab Ha.
rewrite in_tsep;rewrite in_tsep in Ha.
apply ParAlg.up_closed with (¬ a);intuition.
Qed.

Definition tarr:=(@tarrow _ _ _ _ TS).
Infix "⨂↦":=(tarr) (at level 60, right associativity). 

Ltac PAstep ax:= eapply ParAlg.up_closed;[|apply ax]; repeat auto_meet_intro;  auto_meet_elim_trans.

Lemma arrow_tsep_psep: ∀ a b, (tsep ( a ⨂↦ b) ↔ psep ((¬a) ⅋↦ (¬b))).
Proof.
  intros a b ;split;intro H.
  - eapply ParAlg.modus_ponens with ((a⅋ ¬b)).
    + eapply ParAlg.modus_ponens with (¬¬(a⅋ ¬b)).
        * apply in_tsep;assumption.
        * apply dne_entails. 
    + eapply ParAlg.modus_ponens;[exact (dni_entails a)|].
       (*  PAstep sep_PS4_r. 
        instantiate (4:=a). 
        instantiate (1:=¬b). 
        instantiate (1:=¬¬a). reflexivity. *)
        eapply ParAlg.up_closed;[|apply sep_PS4_r].
         apply (meet_elim_trans);later.
        split;[exists a;reflexivity|].
        apply (meet_elim_trans);later.
        split;[exists (¬¬a);reflexivity|].
        apply (meet_elim);exists  (¬b).
        reflexivity.
  -  apply in_tsep.
      eapply ParAlg.modus_ponens with ((a⅋ ¬b)).
    + eapply ParAlg.modus_ponens with ((¬¬a)⅋ ¬b).
        * assumption.
        * eapply ParAlg.modus_ponens;[exact (dne_entails a)|].
           eapply ParAlg.up_closed;[|apply sep_PS4_r].
           apply (meet_elim_trans);later.
           split;[exists (¬¬a);reflexivity|].
           apply (meet_elim_trans);later.
           split;[exists (a);reflexivity|].
          apply (meet_elim);exists  (¬b).
        reflexivity.
    + apply dni_entails.
Qed.

  
Lemma modus_ponens_tsep: ∀ a b, tsep a → tsep (a ⨂↦ b) → tsep b.
Proof.
  intros a b Ha Hab.
  rewrite in_tsep.
  rewrite in_tsep in Ha.
  rewrite arrow_tsep_psep in Hab.
  eapply ParAlg.modus_ponens;[exact Ha|exact Hab].
Qed.

Ltac auto_step_ter p ax:=  eapply sep_PC6_ter with (g:= p);[step ax|].

Ltac auto_rewrite:=
  match goal with
  |[ H: ?x = _ |- ?x = _] => rewrite H; try (reflexivity)
  |[ H: ?x = _ |- ?f ?x = _] => rewrite H; try (reflexivity)
  end.
  
Ltac auto_same_set_step:=
  match goal with
  |[ |- _ ≅ _ ] => let x:=fresh "x" in let a := fresh "a" in let H:=fresh "H" in 
              intro x; unfold image_set; split;intros (a,H)
  |[ H: (∃ _: X, _ = _ ) ∧ ?x = _ |- ∃ _:X, ?x = _] => 
    let b:=fresh "a" in let Ha:=fresh "H" in let Hb:=fresh "H" in
    destruct H as ((b,Hb),Ha);try(later; rewrite Ha,Hb) (* ,neg_join in ;try(now reflexivity) *)
  |[ H: ?x = _ |- ∃ _ : X, ((∃ _ : X, _ = _) ∧ ?x = _ )] => 
      try(later;split;[later;reflexivity|]);try(rewrite H;now reflexivity)
  |[ H: ?x = _ |- ∃ _ : X, ?x = _] => later;repeat(auto_rewrite)
  |[ |- (neg (join_set _)) =  _] => rewrite neg_join
  |[ |- _ = (neg (join_set _))] => rewrite neg_join;repeat(auto_rewrite)
  |[ |- ⊓ _ = ⊓ _] => apply meet_same_set
  |[ |-  _ =  _] => now reflexivity
  end.
Ltac auto_same_set:=repeat (auto_same_set_step).


 
Lemma tsep_TS1: tsep (@TS1 _ _ _ _ TS).
Proof.
  unfold TS1.
  rewrite meet_same_set with (B:=(image_set (@neg _ _ _ _ TS) (λ x : X, ∃ a : X, x= (a ⨂ (¬ (a ⨂ a)))))).
  - rewrite <- neg_join.
    rewrite in_tsep.
    unfold meet_set,tensor,TS,PS_to_TS,psep;simpl.
    eapply ParAlg.modus_ponens;[|exact (dni_entails _)].
    do 2 (auto_add_inf_X X).
    eapply mod_pon_inf_ter with (f:= fun (c b a  :X) => (¬(a⅋a) ⅋ a)).
    + step sep_PS3.
    + step sep_PS1.
  - auto_same_set. 
Qed.

Lemma tsep_TS2: tsep (@TS2 _ _ _ _ TS).
Proof.
  unfold TS2.
  rewrite meet_same_set 
  with (B:=image_set (@neg _ _ _ _ TS) (λ x : X, ∃ a : X, x = @join_set _ _ _ RevCompleteLattice (λ x0 : X, ∃ b : X, x0 =  ((a ⨂ b) ⨂ ¬ a)))).
  - rewrite <- neg_join,in_tsep.
    unfold meet_set,tensor,TS,PS_to_TS,psep;simpl.
    eapply ParAlg.modus_ponens;[|exact (dni_entails _)].
    auto_add_inf_X X.
    eapply mod_pon_inf_ter with (f:= fun (c  a b :X) => ((¬a ) ⅋ (a⅋b))).
    + step sep_PS3.
    + step sep_PS2.
  - auto_same_set.
Qed.

Lemma tsep_TS3: tsep (@TS3 _ _ _ _ TS).
Proof.
  unfold TS3.
  rewrite meet_same_set with (B:=image_set (@neg _ _ _ _ TS) 
      (λ x : X, ∃ a : X, x = @join_set _ _ _ RevCompleteLattice (λ x0 : X, ∃ b : X, x0 =  ((a ⨂ b) ⨂ ¬ (b⨂a))))).
  - rewrite <- neg_join,in_tsep.
    unfold meet_set,tensor,TS,PS_to_TS,psep;simpl.
    eapply ParAlg.modus_ponens;[|exact (dni_entails _)].
    auto_add_inf_X X.
    eapply mod_pon_inf_ter with (f:= fun (c  a b :X) => (b ⅋ a) ⅋↦ (a⅋b)).
    + step sep_PS3.
    + step sep_PS3.
  - auto_same_set.
Qed.

Lemma tsep_TS4: tsep (@TS4 _ _ _ _ TS).
Proof.
  unfold TS4.
  rewrite meet_same_set with (B:=image_set (@neg _ _ _ _ TS) 
      (λ x : X, ∃ a : X, x = @join_set _ _ _ RevCompleteLattice 
      (λ x0 : X, ∃ b : X, x0 =  @join_set _ _ _ RevCompleteLattice
      (λ x1 : X, ∃ c : X, x1 = (a ⅋ ¬ b) ⅋↦ (¬ ¬ ((c ⅋ a) ⅋ ¬ (c ⅋ b))))))).
  - rewrite <- neg_join,in_tsep.
    unfold meet_set,tensor,TS,PS_to_TS,psep;simpl.
    eapply ParAlg.modus_ponens;[|exact (dni_entails _)].
    auto_step_ter (fun (a b c:X) => (b ⅋↦a)) sep_PS3.
    auto_step_ter (fun (a b c:X) => (c ⅋ b) ⅋↦ (c ⅋ a)) sep_PS4.
    auto_step_ter (fun (a b c:X) => (c ⅋ a) ⅋ ¬(c ⅋ b)) sep_PS3.
    step dni.
 - auto_same_set.
Qed.

Lemma tsep_TS5: tsep (@TS5 _ _ _ _ TS).
Proof.
  unfold TS5.
  rewrite meet_same_set with (B:=image_set (@neg _ _ _ _ TS) 
      (λ x : X, ∃ a : X, x = @join_set _ _ _ RevCompleteLattice 
      (λ x0 : X, ∃ b : X, x0 =  @join_set _ _ _ RevCompleteLattice
      (λ x1 : X, ∃ c : X, x1 = ((a ⅋ b ⅋ c) ⅋ ¬ ((a ⅋ b) ⅋ c)))))).
  - rewrite <- neg_join,in_tsep.
    unfold meet_set,tensor,TS,PS_to_TS,psep;simpl.
    eapply ParAlg.modus_ponens;[|exact (dni_entails _)].
    rewrite par_comm_sep.
    step par_assoc_r.
  - auto_same_set.
Qed.
Definition PA_to_TA:=@Build_TensorAlg _ _ _ _ TS tsep modus_ponens_tsep up_closed_tsep
                     tsep_TS1 tsep_TS2 tsep_TS3 tsep_TS4 tsep_TS5.

End ReverseParAlg.


Section About_TensorAlg.
Context `{TA:TensorAlg}.
Notation "¬ a":=(neg a).
Definition abs t:= ⊓(λ x:X, ∃ a:X, x= ⊓ (λ y : X, ∃ b :X, y = (¬(a ⨂ ¬b)) ∧ t a ≤ b)).
Definition antiapp t u:= (⊔(λ x:X, ∃ a:X, x= a ∧ (t ≤ ⊔ ((λ y : X, ∃ b :X, y = (¬ b) ∧ (u ⨂ a) ≤ b))))).
Definition app t u:= (⊓(λ x:X, ∃ a:X, x= a ∧ (t ≤ ⊔ ((λ y : X, ∃ b :X, y = (¬ b) ∧ (u ⨂ ¬a) ≤ b))))).
Notation "λ⨂ t":=(abs t) (at level 60).
Notation "t @⨂ u":=(app t u) (at level 59, left associativity).
Definition eta a:= fun x => a @⨂ x.



Lemma semi_adjunction: ∀ t u r, (¬ r) ≤ antiapp t  u <-> t ≤ u ⨂↦ r.
Proof.
intros t u r;split; intro H.
- apply ord_trans with (¬(u ⨂ antiapp t u)).
  + unfold antiapp in *.
    rewrite tensor_join_r,neg_join.
    apply meet_intro.
    intros x (a,(Ha,Hx)).
    subst x. unfold tensor_set_r in Ha.
    destruct Ha as (b,((c,(Hc,Hd)),Ha)).
    subst a. subst b.
    eapply ord_trans;[exact Hd|].
    apply join_intro.
    intros x (a,(Ha,Hx)).
    subst x.
    intuition.
  + apply neg_mon.
    apply tensor_mon;[reflexivity|assumption].
- apply join_elim.
  later;split;[reflexivity|].
  eapply ord_trans;[exact H|].
  apply join_elim.
  later;split;[reflexivity|].
  reflexivity.
Qed.


(* Lemma preetarule: ∀ t a, t ≤ a ⨂↦ (t@⨂ a).
Proof.
intros t a.
apply semi_adjunction.
unfold antiapp.
apply join_elim.
later;split;[reflexivity|].
apply join_elim_trans. later. split.
- later;split. reflexivity. reflexivity.


Lemma etarule: ∀ a, a ≤ (λ⨂ (eta a)).
intro t.
unfold eta,abs,app.
apply meet_intro.
intros x (a,Ha).
subst x.
apply meet_intro.
intros y (b,(Hy,Hb)).
subst y.
apply ord_trans with (¬(a⨂⊔(image_set neg (λ x : X,∃ a0 : X,
        x = a0 ∧ t ≤ ⊔ (λ y : X, ∃ b : X, y = (¬ b) ∧ (a ⨂ ¬ a0) ≤ b))))).
- rewrite tensor_join_r,neg_join.
  apply meet_intro.
  intros z (c,(Hc,Hz));subst z.
  unfold image_set,tensor_set_r in Hc.
  destruct Hc as (d,((e,((f,(Hfe,Hf)),He)),Hd)).
  subst f. subst d. subst c.
  apply ord_trans with (⊔ (λ y : X, ∃ b : X, y = (¬ b) ∧ (a ⨂ ¬ e) ≤ b));intuition.
  apply join_intro.
  intros x (c,(Hc,Hx));subst x.
  apply neg_mon;assumption.
- apply neg_mon,tensor_mon_r.
  
 intuition.



(* Lemma semi_adjunction_bis: ∀ t u r, t @⨂ u ≤ (r) <-> (¬ r) ≤ antiapp t u.
Proof.
intros t u r;split;intro H.
unfold app,antiapp in *.
-  apply join_elim.
  exists (¬ r);intuition.
  eapply ord_trans.
  Focus 2. apply join_elim.
    later;split;[reflexivity|].
    apply tensor_mon_r,neg_mon,H.
   rewrite neg_join. 
- apply semi_adjunction in H.
  apply meet_elim.
  exists r;intuition.
  eapply ord_trans;[exact H|].
  apply join_elim.
  later;intuition.
Qed.
 *)
Lemma mod_pon_inf: 
   ∀ (f g:X→X→X), separator (⋂[a] ⋂[b] ((f a b) ⨂↦ (g a b))) → separator (⋂[a] ⋂[b] (f a b)) → separator (⋂[a] ⋂[b] (g a b)).
Proof.
  intros f g Hfg Hf.
  remember (⋂[a] ⋂[b] ((f a b) ⨂↦ (g a b))) as Ifg.
  remember (⋂[a] ⋂[b] (f a b)) as If.
  apply up_closed with (Ifg@If);auto.
  repeat auto_meet_intro.
  apply adjunction.
  rewrite HeqIfg.
  auto_meet_elim_trans.
  apply parrow_mon_l.
  rewrite HeqIf.
  auto_meet_elim_trans.
Qed.









 *)        

End About_TensorAlg.